@@ -12,14 +12,12 @@
    | obtain it through the world-wide-web, please send a note to          |
    | license@php.net so we can mail you a copy immediately.               |
    +----------------------------------------------------------------------+
-   | Authors: Andi Gutmans <andi@zend.com>                                |
+   | Authors: Andi Gutmans <andi@php.net>                                 |
    |          Sascha Schumann <sascha@schumann.cx>                        |
    |          Pierre Joye <pierre@php.net>                                |
    +----------------------------------------------------------------------+
 */
 
-/* $Id$ */
-
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <string.h>
@@ -85,14 +83,12 @@ ts_rsrc_id cwd_globals_id;
 virtual_cwd_globals cwd_globals;
 #endif
 
-cwd_state main_cwd_state; /* True global */
+static cwd_state main_cwd_state; /* True global */
 
 #ifndef ZEND_WIN32
 #include <unistd.h>
 #else
 #include <direct.h>
-#include "zend_globals.h"
-#include "zend_globals_macros.h"
 #endif
 
 #define CWD_STATE_COPY(d, s)				\
@@ -157,7 +153,6 @@ typedef struct {
 static inline time_t FileTimeToUnixTime(const FILETIME *FileTime)
 {
 	__int64 UnixTime;
-	long *nsec = NULL;
 	SYSTEMTIME SystemTime;
 	FileTimeToSystemTime(FileTime, &SystemTime);
 
@@ -166,10 +161,6 @@ static inline time_t FileTimeToUnixTime(const FILETIME *FileTime)
 
 	UnixTime -= (SECS_BETWEEN_EPOCHS * SECS_TO_100NS);
 
-	if (nsec) {
-		*nsec = (UnixTime % SECS_TO_100NS) * (__int64)100;
-	}
-
 	UnixTime /= SECS_TO_100NS; /* now convert to seconds */
 
 	if ((time_t)UnixTime != UnixTime) {
@@ -178,7 +169,7 @@ static inline time_t FileTimeToUnixTime(const FILETIME *FileTime)
 	return (time_t)UnixTime;
 }
 
-CWD_API int php_sys_readlink(const char *link, char *target, size_t target_len){ /* {{{ */
+CWD_API ssize_t php_sys_readlink(const char *link, char *target, size_t target_len){ /* {{{ */
 	HANDLE hFile;
 	wchar_t *linkw = php_win32_ioutil_any_to_w(link), targetw[MAXPATHLEN];
 	size_t ret_len, targetw_len, offset = 0;
@@ -194,8 +185,8 @@ CWD_API int php_sys_readlink(const char *link, char *target, size_t target_len){
 	}
 
 	hFile = CreateFileW(linkw,            // file to open
-				 GENERIC_READ,          // open for reading
-				 FILE_SHARE_READ,       // share for reading
+				 0,  // query possible attributes
+				 PHP_WIN32_IOUTIL_DEFAULT_SHARE_MODE,
 				 NULL,                  // default security
 				 OPEN_EXISTING,         // existing file only
 				 FILE_FLAG_BACKUP_SEMANTICS, // normal file
@@ -244,7 +235,7 @@ CWD_API int php_sys_readlink(const char *link, char *target, size_t target_len){
 	CloseHandle(hFile);
 	free(linkw);
 
-	return ret_len;
+	return (ssize_t)ret_len;
 }
 /* }}} */
 
@@ -306,7 +297,13 @@ CWD_API int php_sys_stat_ex(const char *path, zend_stat_t *buf, int lstat) /* {{
 		REPARSE_DATA_BUFFER * pbuffer;
 		DWORD retlength = 0;
 
-		hLink = CreateFileW(pathw, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, NULL);
+		hLink = CreateFileW(pathw,
+				FILE_READ_ATTRIBUTES,
+				PHP_WIN32_IOUTIL_DEFAULT_SHARE_MODE,
+				NULL,
+				OPEN_EXISTING,
+				FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS,
+				NULL);
 		if(hLink == INVALID_HANDLE_VALUE) {
 			free(pathw);
 			return -1;
@@ -424,7 +421,7 @@ void virtual_cwd_main_cwd_init(uint8_t reinit) /* {{{ */
 		cwd[0] = '\0';
 	}
 
-	main_cwd_state.cwd_length = (int)strlen(cwd);
+	main_cwd_state.cwd_length = strlen(cwd);
 #ifdef ZEND_WIN32
 	if (main_cwd_state.cwd_length >= 2 && cwd[1] == ':') {
 		cwd[0] = toupper(cwd[0]);
@@ -731,10 +728,10 @@ CWD_API realpath_cache_bucket** realpath_cache_get_buckets(void)
 #undef LINK_MAX
 #define LINK_MAX 32
 
-static int tsrm_realpath_r(char *path, int start, int len, int *ll, time_t *t, int use_realpath, int is_dir, int *link_is_dir) /* {{{ */
+static size_t tsrm_realpath_r(char *path, size_t start, size_t len, int *ll, time_t *t, int use_realpath, int is_dir, int *link_is_dir) /* {{{ */
 {
-	int i, j, save;
-	int directory = 0;
+	size_t i, j;
+	int directory = 0, save;
 #ifdef ZEND_WIN32
 	WIN32_FIND_DATAW dataw;
 	HANDLE hFind = INVALID_HANDLE_VALUE;
@@ -762,28 +759,31 @@ static int tsrm_realpath_r(char *path, int start, int len, int *ll, time_t *t, i
 		while (i > start && !IS_SLASH(path[i-1])) {
 			i--;
 		}
+		assert(i < MAXPATHLEN);
 
 		if (i == len ||
-			(i == len - 1 && path[i] == '.')) {
+			(i + 1 == len && path[i] == '.')) {
 			/* remove double slashes and '.' */
-			len = i - 1;
+			len = EXPECTED(i > 0) ? i - 1 : 0;
 			is_dir = 1;
 			continue;
-		} else if (i == len - 2 && path[i] == '.' && path[i+1] == '.') {
+		} else if (i + 2 == len && path[i] == '.' && path[i+1] == '.') {
 			/* remove '..' and previous directory */
 			is_dir = 1;
 			if (link_is_dir) {
 				*link_is_dir = 1;
 			}
-			if (i - 1 <= start) {
+			if (i <= start + 1) {
 				return start ? start : len;
 			}
 			j = tsrm_realpath_r(path, start, i-1, ll, t, use_realpath, 1, NULL);
-			if (j > start) {
+			if (j > start && j != (size_t)-1) {
 				j--;
+				assert(i < MAXPATHLEN);
 				while (j > start && !IS_SLASH(path[j])) {
 					j--;
 				}
+				assert(i < MAXPATHLEN);
 				if (!start) {
 					/* leading '..' must not be removed in case of relative path */
 					if (j == 0 && path[0] == '.' && path[1] == '.' &&
@@ -823,7 +823,7 @@ static int tsrm_realpath_r(char *path, int start, int len, int *ll, time_t *t, i
 			if ((bucket = realpath_cache_find(path, len, *t)) != NULL) {
 				if (is_dir && !bucket->is_dir) {
 					/* not a directory */
-					return -1;
+					return (size_t)-1;
 				} else {
 					if (link_is_dir) {
 						*link_is_dir = bucket->is_dir;
@@ -838,14 +838,14 @@ static int tsrm_realpath_r(char *path, int start, int len, int *ll, time_t *t, i
 		if (save) {
 			pathw = php_win32_ioutil_any_to_w(path);
 			if (!pathw) {
-				return -1;
+				return (size_t)-1;
 			}
 			hFind = FindFirstFileExW(pathw, FindExInfoBasic, &dataw, FindExSearchNameMatch, NULL, 0);
 			if (INVALID_HANDLE_VALUE == hFind) {
 				if (use_realpath == CWD_REALPATH) {
 					/* file not found */
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 				/* continue resolution anyway but don't save result in the cache */
 				save = 0;
@@ -857,7 +857,6 @@ static int tsrm_realpath_r(char *path, int start, int len, int *ll, time_t *t, i
 		tmp = do_alloca(len+1, use_heap);
 		memcpy(tmp, path, len+1);
 
-retry:
 		if(save &&
 				!(IS_UNC_PATH(path, len) && len >= 3 && path[2] != '?') &&
                                (dataw.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT)
@@ -866,7 +865,8 @@ retry:
 			HANDLE hLink = NULL;
 			REPARSE_DATA_BUFFER * pbuffer;
 			DWORD retlength = 0;
-			int bufindex = 0, isabsolute = 0;
+			size_t bufindex = 0;
+			uint8_t isabsolute = 0;
 			wchar_t * reparsetarget;
 			BOOL isVolume = FALSE;
 #if VIRTUAL_CWD_DEBUG
@@ -874,20 +874,26 @@ retry:
 #endif
 			char *substitutename = NULL;
 			size_t substitutename_len;
-			int substitutename_off = 0;
+			size_t substitutename_off = 0;
 			wchar_t tmpsubstname[MAXPATHLEN];
 
 			if(++(*ll) > LINK_MAX) {
 				free_alloca(tmp, use_heap);
 				FREE_PATHW()
-				return -1;
+				return (size_t)-1;
 			}
 
-			hLink = CreateFileW(pathw, 0, 0, NULL, OPEN_EXISTING, FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS, NULL);
+			hLink = CreateFileW(pathw,
+					0,
+					PHP_WIN32_IOUTIL_DEFAULT_SHARE_MODE,
+					NULL,
+					OPEN_EXISTING,
+					FILE_FLAG_OPEN_REPARSE_POINT|FILE_FLAG_BACKUP_SEMANTICS,
+					NULL);
 			if(hLink == INVALID_HANDLE_VALUE) {
 				free_alloca(tmp, use_heap);
 				FREE_PATHW()
-				return -1;
+				return (size_t)-1;
 			}
 
 			pbuffer = (REPARSE_DATA_BUFFER *)do_alloca(MAXIMUM_REPARSE_DATA_BUFFER_SIZE, use_heap_large);
@@ -895,28 +901,14 @@ retry:
 				CloseHandle(hLink);
 				free_alloca(tmp, use_heap);
 				FREE_PATHW()
-				return -1;
+				return (size_t)-1;
 			}
 			if(!DeviceIoControl(hLink, FSCTL_GET_REPARSE_POINT, NULL, 0, pbuffer,  MAXIMUM_REPARSE_DATA_BUFFER_SIZE, &retlength, NULL)) {
-				BY_HANDLE_FILE_INFORMATION fileInformation;
-
 				free_alloca(pbuffer, use_heap_large);
-				if ((dataw.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) &&
-						(dataw.dwReserved0 & ~IO_REPARSE_TAG_CLOUD_MASK) == IO_REPARSE_TAG_CLOUD &&
-						EG(windows_version_info).dwMajorVersion >= 10 &&
-						EG(windows_version_info).dwMinorVersion == 0 &&
-						EG(windows_version_info).dwBuildNumber >= 18362 &&
-						GetFileInformationByHandle(hLink, &fileInformation) &&
-						!(fileInformation.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT)) {
-					dataw.dwFileAttributes = fileInformation.dwFileAttributes;
-					CloseHandle(hLink);
-					(*ll)--;
-					goto retry;
-				}
 				free_alloca(tmp, use_heap);
 				CloseHandle(hLink);
 				FREE_PATHW()
-				return -1;
+				return (size_t)-1;
 			}
 
 			CloseHandle(hLink);
@@ -930,7 +922,7 @@ retry:
 					free_alloca(pbuffer, use_heap_large);
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 #endif
 
@@ -939,7 +931,7 @@ retry:
 					free_alloca(pbuffer, use_heap_large);
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 				memmove(tmpsubstname, reparsetarget + pbuffer->MountPointReparseBuffer.SubstituteNameOffset / sizeof(WCHAR), pbuffer->MountPointReparseBuffer.SubstituteNameLength);
 				tmpsubstname[substitutename_len] = L'\0';
@@ -952,7 +944,7 @@ retry:
 					free(printname);
 #endif
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 			}
 			else if(pbuffer->ReparseTag == IO_REPARSE_TAG_MOUNT_POINT) {
@@ -964,7 +956,7 @@ retry:
 					free_alloca(pbuffer, use_heap_large);
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 #endif
 
@@ -974,7 +966,7 @@ retry:
 					free_alloca(pbuffer, use_heap_large);
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 				memmove(tmpsubstname, reparsetarget + pbuffer->MountPointReparseBuffer.SubstituteNameOffset / sizeof(WCHAR), pbuffer->MountPointReparseBuffer.SubstituteNameLength);
 				tmpsubstname[substitutename_len] = L'\0';
@@ -987,12 +979,13 @@ retry:
 					free(printname);
 #endif
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 			}
 			else if (pbuffer->ReparseTag == IO_REPARSE_TAG_DEDUP ||
 					/* Starting with 1709. */
-					(pbuffer->ReparseTag & ~IO_REPARSE_TAG_CLOUD_MASK) == IO_REPARSE_TAG_CLOUD ||
+					(pbuffer->ReparseTag & IO_REPARSE_TAG_CLOUD_MASK) != 0 && 0x90001018L != pbuffer->ReparseTag ||
+					IO_REPARSE_TAG_CLOUD == pbuffer->ReparseTag ||
 					IO_REPARSE_TAG_ONEDRIVE == pbuffer->ReparseTag) {
 				isabsolute = 1;
 				substitutename = malloc((len + 1) * sizeof(char));
@@ -1000,7 +993,7 @@ retry:
 					free_alloca(pbuffer, use_heap_large);
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 				memcpy(substitutename, path, len + 1);
 				substitutename_len = len;
@@ -1009,7 +1002,7 @@ retry:
 				free_alloca(pbuffer, use_heap_large);
 				free_alloca(tmp, use_heap);
 				FREE_PATHW()
-				return -1;
+				return (size_t)-1;
 			}
 
 			if(isabsolute && substitutename_len > 4) {
@@ -1036,7 +1029,7 @@ retry:
 
 			if (!isVolume) {
 				char * tmp2 = substitutename + substitutename_off;
-				for(bufindex = 0; bufindex < (substitutename_len - substitutename_off); bufindex++) {
+				for (bufindex = 0; bufindex + substitutename_off < substitutename_len; bufindex++) {
 					*(path + bufindex) = *(tmp2 + bufindex);
 				}
 
@@ -1060,10 +1053,10 @@ retry:
 				if (!((j == 3) && (path[1] == ':') && (path[2] == '\\'))) {
 					/* use_realpath is 0 in the call below coz path is absolute*/
 					j = tsrm_realpath_r(path, 0, j, ll, t, 0, is_dir, &directory);
-					if(j < 0) {
+					if(j == (size_t)-1) {
 						free_alloca(tmp, use_heap);
 						FREE_PATHW()
-						return -1;
+						return (size_t)-1;
 					}
 				}
 			}
@@ -1071,17 +1064,17 @@ retry:
 				if(i + j >= MAXPATHLEN - 1) {
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 
 				memmove(path+i, path, j+1);
 				memcpy(path, tmp, i-1);
 				path[i-1] = DEFAULT_SLASH;
 				j  = tsrm_realpath_r(path, start, i + j, ll, t, use_realpath, is_dir, &directory);
-				if(j < 0) {
+				if(j == (size_t)-1) {
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 			}
 			directory = (dataw.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);
@@ -1097,14 +1090,14 @@ retry:
 					/* not a directory */
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
 			}
 #else
 		if (save && php_sys_lstat(path, &st) < 0) {
 			if (use_realpath == CWD_REALPATH) {
 				/* file not found */
-				return -1;
+				return (size_t)-1;
 			}
 			/* continue resolution anyway but don't save result in the cache */
 			save = 0;
@@ -1114,30 +1107,30 @@ retry:
 		memcpy(tmp, path, len+1);
 
 		if (save && S_ISLNK(st.st_mode)) {
-			if (++(*ll) > LINK_MAX || (j = php_sys_readlink(tmp, path, MAXPATHLEN)) < 0) {
+			if (++(*ll) > LINK_MAX || (j = (size_t)php_sys_readlink(tmp, path, MAXPATHLEN)) == (size_t)-1) {
 				/* too many links or broken symlinks */
 				free_alloca(tmp, use_heap);
-				return -1;
+				return (size_t)-1;
 			}
 			path[j] = 0;
 			if (IS_ABSOLUTE_PATH(path, j)) {
 				j = tsrm_realpath_r(path, 1, j, ll, t, use_realpath, is_dir, &directory);
-				if (j < 0) {
+				if (j == (size_t)-1) {
 					free_alloca(tmp, use_heap);
-					return -1;
+					return (size_t)-1;
 				}
 			} else {
 				if (i + j >= MAXPATHLEN-1) {
 					free_alloca(tmp, use_heap);
-					return -1; /* buffer overflow */
+					return (size_t)-1; /* buffer overflow */
 				}
 				memmove(path+i, path, j+1);
 				memcpy(path, tmp, i-1);
 				path[i-1] = DEFAULT_SLASH;
 				j = tsrm_realpath_r(path, start, i + j, ll, t, use_realpath, is_dir, &directory);
-				if (j < 0) {
+				if (j == (size_t)-1) {
 					free_alloca(tmp, use_heap);
-					return -1;
+					return (size_t)-1;
 				}
 			}
 			if (link_is_dir) {
@@ -1152,24 +1145,24 @@ retry:
 				if (is_dir && !directory) {
 					/* not a directory */
 					free_alloca(tmp, use_heap);
-					return -1;
+					return (size_t)-1;
 				}
 			}
 #endif
-			if (i - 1 <= start) {
+			if (i <= start + 1) {
 				j = start;
 			} else {
 				/* some leading directories may be unaccessable */
 				j = tsrm_realpath_r(path, start, i-1, ll, t, save ? CWD_FILEPATH : use_realpath, 1, NULL);
-				if (j > start) {
+				if (j > start && j != (size_t)-1) {
 					path[j++] = DEFAULT_SLASH;
 				}
 			}
 #ifdef ZEND_WIN32
-			if (j < 0 || j + len - i >= MAXPATHLEN-1) {
+			if (j == (size_t)-1 || j + len >= MAXPATHLEN - 1 + i) {
 				free_alloca(tmp, use_heap);
 				FREE_PATHW()
-				return -1;
+				return (size_t)-1;
 			}
 			if (save) {
 				size_t sz;
@@ -1177,9 +1170,9 @@ retry:
 				if (!tmp_path) {
 					free_alloca(tmp, use_heap);
 					FREE_PATHW()
-					return -1;
+					return (size_t)-1;
 				}
-				i = (int)sz;
+				i = sz;
 				memcpy(path+j, tmp_path, i+1);
 				free(tmp_path);
 				j += i;
@@ -1190,9 +1183,9 @@ retry:
 			}
 		}
 #else
-			if (j < 0 || j + len - i >= MAXPATHLEN-1) {
+			if (j == (size_t)-1 || j + len >= MAXPATHLEN - 1 + i) {
 				free_alloca(tmp, use_heap);
-				return -1;
+				return (size_t)-1;
 			}
 			memcpy(path+j, tmp+i, len-i+1);
 			j += (len-i);
@@ -1218,16 +1211,16 @@ retry:
 /* returns 0 for ok, 1 for error */
 CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func verify_path, int use_realpath) /* {{{ */
 {
-	int path_length = (int)strlen(path);
-	char resolved_path[MAXPATHLEN];
-	int start = 1;
+	size_t path_length = strlen(path);
+	char resolved_path[MAXPATHLEN] = {0};
+	size_t start = 1;
 	int ll = 0;
 	time_t t;
 	int ret;
 	int add_slash;
 	void *tmp;
 
-	if (path_length <= 0 || path_length >= MAXPATHLEN-1) {
+	if (!path_length || path_length >= MAXPATHLEN-1) {
 #ifdef ZEND_WIN32
 		_set_errno(EINVAL);
 #else
@@ -1249,7 +1242,7 @@ CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func
 			start = 0;
 			memcpy(resolved_path , path, path_length + 1);
 		} else {
-			int state_cwd_length = state->cwd_length;
+			size_t state_cwd_length = state->cwd_length;
 
 #ifdef ZEND_WIN32
 			if (IS_SLASH(path[0])) {
@@ -1343,7 +1336,7 @@ CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func
 	t = CWDG(realpath_cache_ttl) ? 0 : -1;
 	path_length = tsrm_realpath_r(resolved_path, start, path_length, &ll, &t, use_realpath, 0, NULL);
 
-	if (path_length < 0) {
+	if (path_length == (size_t)-1) {
 		errno = ENOENT;
 		return 1;
 	}
@@ -1351,6 +1344,7 @@ CWD_API int virtual_file_ex(cwd_state *state, const char *path, verify_path_func
 	if (!start && !path_length) {
 		resolved_path[path_length++] = '.';
 	}
+
 	if (add_slash && path_length && !IS_SLASH(resolved_path[path_length-1])) {
 		if (path_length >= MAXPATHLEN-1) {
 			return -1;
@@ -1404,7 +1398,7 @@ CWD_API int virtual_chdir(const char *path) /* {{{ */
 
 CWD_API int virtual_chdir_file(const char *path, int (*p_chdir)(const char *path)) /* {{{ */
 {
-	int length = (int)strlen(path);
+	size_t length = strlen(path);
 	char *temp;
 	int retval;
 	ALLOCA_FLAG(use_heap)
@@ -1412,10 +1406,10 @@ CWD_API int virtual_chdir_file(const char *path, int (*p_chdir)(const char *path
 	if (length == 0) {
 		return 1; /* Can't cd to empty string */
 	}
-	while(--length >= 0 && !IS_SLASH(path[length])) {
+	while(--length < SIZE_MAX && !IS_SLASH(path[length])) {
 	}
 
-	if (length == -1) {
+	if (length == SIZE_MAX) {
 		/* No directory only file name */
 		errno = ENOENT;
 		return -1;
@@ -1459,7 +1453,7 @@ CWD_API char *virtual_realpath(const char *path, char *real_path) /* {{{ */
 	}
 
 	if (virtual_file_ex(&new_state, path, NULL, CWD_REALPATH)==0) {
-		int len = new_state.cwd_length>MAXPATHLEN-1?MAXPATHLEN-1:new_state.cwd_length;
+		size_t len = new_state.cwd_length>MAXPATHLEN-1?MAXPATHLEN-1:new_state.cwd_length;
 
 		memcpy(real_path, new_state.cwd, len);
 		real_path[len] = '\0';
@@ -1923,7 +1917,7 @@ CWD_API char *tsrm_realpath(const char *path, char *real_path) /* {{{ */
 	} else if (!IS_ABSOLUTE_PATH(path, strlen(path)) &&
 					VCWD_GETCWD(cwd, MAXPATHLEN)) {
 		new_state.cwd = estrdup(cwd);
-		new_state.cwd_length = (int)strlen(cwd);
+		new_state.cwd_length = strlen(cwd);
 	} else {
 		new_state.cwd = (char*)emalloc(1);
 		new_state.cwd[0] = '\0';
@@ -1936,7 +1930,7 @@ CWD_API char *tsrm_realpath(const char *path, char *real_path) /* {{{ */
 	}
 
 	if (real_path) {
-		int copy_len = new_state.cwd_length>MAXPATHLEN-1 ? MAXPATHLEN-1 : new_state.cwd_length;
+		size_t copy_len = new_state.cwd_length>MAXPATHLEN-1 ? MAXPATHLEN-1 : new_state.cwd_length;
 		memcpy(real_path, new_state.cwd, copy_len);
 		real_path[copy_len] = '\0';
 		efree(new_state.cwd);
