@@ -17,8 +17,6 @@
    +----------------------------------------------------------------------+
 */
 
-/* $Id$ */
-
 #ifndef _PHP_CURL_H
 #define _PHP_CURL_H
 
@@ -99,15 +97,11 @@ PHP_FUNCTION(curl_share_init);
 PHP_FUNCTION(curl_share_setopt);
 PHP_FUNCTION(curl_share_errno);
 
-#if LIBCURL_VERSION_NUM >= 0x070c00 /* 7.12.0 */
 PHP_FUNCTION(curl_strerror);
 PHP_FUNCTION(curl_multi_strerror);
 PHP_FUNCTION(curl_share_strerror);
-#endif
 
-#if LIBCURL_VERSION_NUM >= 0x070c01 /* 7.12.1 */
 PHP_FUNCTION(curl_reset);
-#endif
 
 #if LIBCURL_VERSION_NUM >= 0x070f04 /* 7.15.4 */
 PHP_FUNCTION(curl_escape);
@@ -154,9 +148,6 @@ typedef struct {
 	php_curl_write    *write;
 	php_curl_write    *write_header;
 	php_curl_read     *read;
-#if CURLOPT_PASSWDFUNCTION != 0
-	zval               passwd;
-#endif
 	zval               std_err;
 	php_curl_progress *progress;
 #if LIBCURL_VERSION_NUM >= 0x071500 /* Available since 7.21.0 */
