@@ -233,7 +233,7 @@ void intl_errors_set_code( intl_error* err, UErrorCode err_code )
 void intl_register_IntlException_class( void )
 {
 	zend_class_entry ce;
-	
+
 	/* Create and register 'IntlException' class. */
 	INIT_CLASS_ENTRY_EX( ce, "IntlException", sizeof( "IntlException" ) - 1, NULL );
 	IntlException_ce_ptr = zend_register_internal_class_ex( &ce,
@@ -280,7 +280,7 @@ smart_str intl_parse_error_to_string( UParseError* pe )
 		}
 		else {
 			smart_str_append( &ret, u8str );
-			zend_string_release( u8str );
+			zend_string_release_ex( u8str, 0 );
 		}
 		smart_str_appends( &ret, "\"" );
 		any = 1;
@@ -300,7 +300,7 @@ smart_str intl_parse_error_to_string( UParseError* pe )
 		else
 		{
 			smart_str_append( &ret, u8str );
-			zend_string_release( u8str );
+			zend_string_release_ex( u8str, 0 );
 		}
 		smart_str_appends( &ret, "\"" );
 		any = 1;
