@@ -18,8 +18,6 @@
    +----------------------------------------------------------------------+
 */
 
-/* $Id$ */
-
 #include "php.h"
 #include "php_globals.h"
 #include "ext/standard/basic_functions.h"
@@ -139,11 +137,10 @@ static void userfilter_dtor(php_stream_filter *thisfilter)
 
 	ZVAL_STRINGL(&func_name, "onclose", sizeof("onclose")-1);
 
-	call_user_function_ex(NULL,
+	call_user_function(NULL,
 			obj,
 			&func_name,
 			&retval,
-			0, NULL,
 			0, NULL);
 
 	zval_ptr_dtor(&retval);
@@ -180,7 +177,7 @@ php_stream_filter_status_t userfilter_filter(
 
 		/* Give the userfilter class a hook back to the stream */
 		php_stream_to_zval(stream, &tmp);
-		zval_copy_ctor(&tmp);
+		Z_ADDREF(tmp);
 		add_property_zval(obj, "stream", &tmp);
 		/* add_property_zval increments the refcount which is unwanted here */
 		zval_ptr_dtor(&tmp);
@@ -254,7 +251,7 @@ php_stream_filter_status_t userfilter_filter(
 	return ret;
 }
 
-static php_stream_filter_ops userfilter_ops = {
+static const php_stream_filter_ops userfilter_ops = {
 	userfilter_filter,
 	userfilter_dtor,
 	"user-filter"
@@ -344,11 +341,10 @@ static php_stream_filter *user_filter_factory_create(const char *filtername,
 	/* invoke the constructor */
 	ZVAL_STRINGL(&func_name, "oncreate", sizeof("oncreate")-1);
 
-	call_user_function_ex(NULL,
+	call_user_function(NULL,
 			&obj,
 			&func_name,
 			&retval,
-			0, NULL,
 			0, NULL);
 
 	if (Z_TYPE(retval) != IS_UNDEF) {
@@ -380,14 +376,14 @@ static php_stream_filter *user_filter_factory_create(const char *filtername,
 	return filter;
 }
 
-static php_stream_filter_factory user_filter_factory = {
+static const php_stream_filter_factory user_filter_factory = {
 	user_filter_factory_create
 };
 
 static void filter_item_dtor(zval *zv)
 {
 	struct php_user_filter_data *fdat = Z_PTR_P(zv);
-	zend_string_release(fdat->classname);
+	zend_string_release_ex(fdat->classname, 0);
 	efree(fdat);
 }
 
@@ -587,10 +583,10 @@ PHP_FUNCTION(stream_filter_register)
 	fdat->classname = zend_string_copy(classname);
 
 	if (zend_hash_add_ptr(BG(user_filter_map), filtername, fdat) != NULL &&
-			php_stream_filter_register_factory_volatile(ZSTR_VAL(filtername), &user_filter_factory) == SUCCESS) {
+			php_stream_filter_register_factory_volatile(filtername, &user_filter_factory) == SUCCESS) {
 		RETVAL_TRUE;
 	} else {
-		zend_string_release(classname);
+		zend_string_release_ex(classname, 0);
 		efree(fdat);
 	}
 }
